/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/module_application.h"
#include "common/software_version.h"
#include <stdio.h>
#include <boost/algorithm/string.hpp>
#include <regex>

// ##################################################################
/**
 * @brief Initialisieren der Membervariablen.
 *
 * @param id Nummer der Kanalklasse fuer z.B. Trace.
 */
Datafox::Communication::Module_application::Module_application( int connection_id, uint8_t device_address )
    : Module_system( connection_id, device_address ) 
{
    trace( "%s(%i):INFO(3): %s [ %i, %u ]\n", __FILE__, __LINE__, __func__, connection_id, device_address );

    // Anzahl anstehender Datensaetze
    std::srand( (unsigned int)time( nullptr ) );

    // Callback fuer Nachricht auf Datensaetze
    m_call_back_notify_record_available = nullptr;
}

// ##################################################################
/**
 * @brief Destruktor.
 *
 */
Datafox::Communication::Module_application::~Module_application() 
{
    while( m_records.size() ) 
    {
        m_records.pop();
    }
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void Datafox::Communication::Module_application::error_module( uint16_t module, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Ist IV_MODULE_APPLICATION ?
    if ( module == IV_MODULE_APPLICATION )
    {

    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    Module_system::error_module( module, error, text );
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void Datafox::Communication::Module_application::error_function( uint16_t module, uint16_t func, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    if ( module == IV_MODULE_APPLICATION )
    {
        // Sonderfehler für Passwortschutz abhandeln.
        if (error == 28) 
        {
            // Die Modulfunktionen liefern den Fehlercode 28, wenn kein oder ein falsches Passwort verwendet wurde.
            // Es ist in diesem Fall zuerst das richtige Passwort für die Kommunikation erforderlich.
            text = boost::str( boost::format( "( %u ): Korrektes Passwort für die Kommunikation erforderlich.") % error );
            return;
        }
        else if (error == 29) 
        {
            // Die Modulfunktionen liefern den Fehlercode 29, wenn 3 mal ein falsches Passwort verwendet wurde.
            // Die Befehlsausführung ist danach für 60 Sekunden blockiert.
            text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
            return;
        }

        switch (func)
        {
        case IV_FUNCTION_CLEAR_RECORDS:
        case IV_FUNCTION_GET_RECORD:
        case IV_FUNCTION_SET_RECORD:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): ") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_SHOW_MESSAGE:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): Nachrichtentext zu lang fuer die Anzeige.") % error ); break;
            case 2:text = boost::str( boost::format( "( %u ): Es wird bereits eine Nachricht angezeigt.") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_SET_VARIABLE:
        case IV_FUNCTION_GET_VARIABLE:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): Der Majorbestandteil des Variablennamens ist unbekannt.") % error ); break;
            case 2:text = boost::str( boost::format( "( %u ): Der Minorbestandteil des Variablennamens ist unbekannt.") % error ); break;
            case 3:
                text = boost::str( boost::format( "( %u ): Der verwendete Variablenwert ist zu lang.") % error );
                break;

            case 4:
                text = boost::str( boost::format( "( %u ): Der verwendete Variablentype würde nicht eingehalten.") % error );
                break;

            case 5:
                text = boost::str( boost::format( "( %u ): Das Schreib-/Leserecht würde nicht eingehalten.") % error );
                break;

            case 6:
                text = boost::str( boost::format( "( %u ): Das verwendete Format oder der Wertebereich würde nicht eingehalten.") % error );
                break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_SHOW_INFO:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): Nachrichtentext zu lang fuer die Anzeige.") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND:
        case IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK:
            switch(error) {
            case 1: text = boost::str( boost::format( "( %u ): ZK nicht aktiv.") % error ); break;
            case 2: text = boost::str( boost::format( "( %u ): Parameter Command außerhalb seines gültigen Bereichs.") % error ); break;
            case 3: text = boost::str( boost::format( "( %u ): Parameter Master ID außerhalb seines gültigen Bereichs.") % error ); break;
            case 4: text = boost::str( boost::format( "( %u ): Parameter KnobId außerhalb seines gültigen Bereichs.") % error ); break;
            case 5: text = boost::str( boost::format( "( %u ): Speicherfehler.") % error ); break;
            case 6: text = boost::str( boost::format( "( %u ): Kommunikationsfehler mit dem Funkknoten.") % error ); break;
            case 7: text = boost::str( boost::format( "( %u ): Die Ausführung des Befehls erfolgte, doch der Schließlinder konnte nicht entfernt werden.") % error ); break;
            case 8: text = boost::str( boost::format( "( %u ): Die Ausführung des Befehls erfolgte, doch der Schließlinder konnte nicht eingelernt werden.") % error ); break;
            case 9: text = boost::str( boost::format( "( %u ): Die Ausführung des Befehls erfolgte, doch der Schließlinder konnte nicht eingelernt werden, weil er einen Produktschutz hat.") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Fehlerblock für Funktion!") % error ); break;
        }
        
        return;
    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    Module_system::error_function( module, func, error, text );
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::remote_call( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count )
{
    if (module == IV_MODULE_APPLICATION)
    {
        unsigned char status;
        switch (func)
        {
        case IV_FUNCTION_CLEAR_RECORDS: return clear_records(get_device_address_sender());
        case IV_FUNCTION_SET_RECORD: return quit_record(get_device_address_sender(), status);
        case IV_FUNCTION_GET_RECORD: return read_record(get_device_address_sender(), data, (std::size_t &)count, status);
        case IV_FUNCTION_SHOW_MESSAGE: return write_message_to_display(get_device_address_sender(), data[0], data[1], data[2], (char*)&data[3]);
        case IV_FUNCTION_SET_VARIABLE:
        {
            std::string name(reinterpret_cast<char *>(&data[0]));
            std::string val(reinterpret_cast<char *>(&data[name.size() + 1]));
            return write_variable(get_device_address_sender(), name, val);
        }
        case IV_FUNCTION_GET_VARIABLE:
        {
            std::string name(reinterpret_cast<char *>(&data[0]));
            std::string val;
            int res = read_variable(get_device_address_sender(), name, val);
            if (res == IV_ERROR_SUCCESS)
            {
                std::copy(val.cbegin(), val.cend(), &data[name.size() + 1]);
                count = static_cast<int>( name.size() + 1 + val.size() + 1 );
            }

            return res;
        }
        case IV_FUNCTION_SHOW_INFO: return write_memo_to_display(get_device_address_sender(), data[0], data[1], data[2], data[3], *(short*)&data[6], (char*)&data[8]);
        case IV_FUNCTION_ENTRANCE2_IDENTIFICATION: 
            {
            char *id, *pin;
            char status[64]; 
            int statusLength;
            id = (char*)&data[strlen((char*)&data[0]) + 1];
            pin = &id[strlen(id) + 1];
            return access_control_identification(get_device_address_sender(), (char*)&data[0], id, pin, status, statusLength); 
            }
        case IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK: return reply_access_control_identification(get_device_address_sender());
        case IV_FUNCTION_NOTIFY: return notify(get_device_address_sender(), data[0], *(short*)&data[1], (char*)&data[3], (char*)&data[3 + strlen((char*)&data[3])]);
        case IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK: return reply_access_control_identification_ext( get_device_address_sender() );
        case IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK: return reply_access_control_knob_command( get_device_address_sender() );
        case IV_FUNCTION_CUSTOM_COMMAND_OK: return reply_custom_command( get_device_address_sender() );
		// Funktion nicht definiert
        default: return 0x0200; 
        }

    }

    return Module_system::remote_call(module, func, data, count);
}

//#########################################################
//##################### APPLICATION #######################
//#########################################################

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::clear_records( unsigned char device_address )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address );

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        get_protocol().set_device_address_receiver( device_address );

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_CLEAR_RECORDS, m_packet_data, count )) != IV_ERROR_SUCCESS ) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_CLEAR_RECORDS, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count ) 
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        while (m_records.size())
        {
            m_records.pop();
        }

        // Antwort des Befehls senden.
        if ( (result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_CLEAR_RECORDS, tmp, count)) != IV_ERROR_SUCCESS )
        {

            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param device_address Anzusprechende Geraetenummer.
*/
int Datafox::Communication::Module_application::restore_records( unsigned char device_address )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_RESTORE_RECORDS, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_RESTORE_RECORDS, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

//		while (m_records.size())
//		{
//			m_records.pop();
//		}

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_RESTORE_RECORDS, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::quit_record( uint8_t device_address, uint8_t &status )
{
    trace("%s(%i):INFO(3): %s [ %u, %u ]\n", __FILE__, __LINE__, __func__, device_address, status);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, result );
            return result;
        }

        // OK
        if (count > (int)sizeof(unsigned char)) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }
        
        switch (m_packet_data[0]) 
        {
        case 0:
        case 2:
            status = 0;
            break;

        case MB_ANSW_ACK_ON_RECORD:        // Quittung erhalten (fuer Online-Datensatz) 
        case MB_ANSW_TIMEOUT_ACK_RECORD:   // Quittung kam nach Timeout (fuer Online-Datensatz) 
            status = 1;
            break;

        case MB_ANSW_OLD_ACK_ON_RECORD:    // wiederholt gesendete Quittung 
        case MB_ANSW_OLD_TIMEOUT_ACK_RECORD: // wiederholte Quittung nach Timeout 
            status = 2;
            break;

        case MB_ANSW_ACK_OFF_RECORD:       // Quittung erhalten (fuer Offline-Datensatz) 
            status = 3;
            break;

        case MB_ANSW_OLD_ACK_OFF_RECORD:   // wiederholt gesendete Quittung 
            status = 4;
            break;
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char data[1];

        if (m_records.size())
        {
            CRecord &record = m_records.front();

            switch (record.m_data[0])
            {
            case MB_ANSW_NO_RECORD:             // kein Datensatz vorhanden
                data[0] = 0;
                break;

            case MB_ANSW_ONLINE_RECORD:         // Online Datensatz
            case MB_ANSW_OLD_ONLINE_RECORD:     // wiederholt gesendeter Online Datensatz
                data[0] = MB_ANSW_ACK_ON_RECORD;
                break;

            case MB_ANSW_OFFLINE_RECORD:        // Offline Datensatz
            case MB_ANSW_OLD_OFFLINE_RECORD:    // wiederholt gesendeter Offline Datensatz
                data[0] = MB_ANSW_ACK_OFF_RECORD;
                break;
            }

            count = sizeof(unsigned char);

            // Datensatz loeschen
            m_records.pop();
        }
        else
        {
            data[0] = 0;
            count = sizeof(unsigned char);
        }

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, data, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::read_record( uint8_t device_address, uint8_t *buffer, std::size_t &count, uint8_t &status )
{
    trace("%s(%i):INFO(3): %s [ %u, %p, %u, %u ]\n", __FILE__, __LINE__, __func__, device_address, buffer, count, status);

    int result = IV_ERROR_SUCCESS;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_RECORD, m_packet_data, (uint16_t &)count) ) != IV_ERROR_SUCCESS ) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_GET_RECORD, result );
            return result;
        }

        // OK
        
        switch ( m_packet_data[0] ) 
        {
        case MB_ANSW_NO_RECORD:             // kein Datensatz vorhanden
            status = 0;
            break;

        case MB_ANSW_ONLINE_RECORD:         // Online Datensatz 
            status = 1;
            break;

        case MB_ANSW_OLD_ONLINE_RECORD:     // wiederholt gesendeter Online Datensatz 
            status = 2;
            break;

        case MB_ANSW_OFFLINE_RECORD:        // Offline Datensatz 
            status = 3;
            break;

        case MB_ANSW_OLD_OFFLINE_RECORD:    // wiederholt gesendeter Offline Datensatz
            status = 4;
            break;
        }
        
        if ( ((status == 0) && (count > 1)) || (count > 243) ) 
        {
            return -IV_ERROR_PROCESSING_DLL;
        }


        for ( uint16_t i = 0; i < count; i++ ) 
        {
            buffer[i] = m_packet_data[i];
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char data[256];
        count = 0;

        if (m_records.size())
        {
            CRecord &record = m_records.front();

            // Daten kopieren
            data[0] = record.m_status;
            memcpy(&data[1], record.m_data, record.m_length);
            count = 1 + record.m_length;
        }
        else
        {
            data[count++] = MB_ANSW_NO_RECORD;
        }

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_RECORD, data, (uint16_t &)count )) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::write_message_to_display( unsigned char device_address, unsigned char time, unsigned char action, unsigned char audio, const std::string &text )
{
    trace("%s(%i):INFO(3): %s [ %u, %u, %u, %u, %p ]\n", __FILE__, __LINE__, __func__, device_address, time, action, audio, text.c_str());

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        m_packet_data[count] = time;
        count += sizeof(unsigned char);
        m_packet_data[count] = action;
        count += sizeof(unsigned char);
        m_packet_data[count] = audio;
        count += sizeof(unsigned char);

        // Text uebernehmen
        std::copy(text.cbegin(), text.cend(), &m_packet_data[count]);
        count += static_cast<int>( text.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_MESSAGE, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_MESSAGE, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_MESSAGE, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, text.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::write_variable( unsigned char device_address, const std::string &name, const std::string &val )
{
    trace("%s(%i):INFO(3): %s [ %i, %s, %s ]\n", __FILE__, __LINE__, __func__, device_address, name.c_str(), val.c_str());

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Namen uebernehmen
        std::copy(name.cbegin(), name.cend(), &m_packet_data[count]);
        count += static_cast<int>( name.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;

        // Wert uebernehmen
        std::copy(val.cbegin(), val.cend(), &m_packet_data[count]);
        count += static_cast<int>( val.size() );
        m_packet_data[count] = 0;
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_VARIABLE, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_SET_VARIABLE, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_VARIABLE, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s = %s ]\n", __FILE__, __LINE__, __func__, name.c_str(), val.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::read_variable( unsigned char device_address, const std::string &name, std::string &val )
{
    trace("%s(%i):INFO(3): %s [ %i, %s, 0x%08X ]\n", __FILE__, __LINE__, __func__, device_address, name.c_str(), val.c_str());

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Namen uebernehmen
        std::copy(name.cbegin(), name.cend(), &m_packet_data[count]);
        count += static_cast<int>( name.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_VARIABLE, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_GET_VARIABLE, result );
            return result;
        }

        // Name+Wert als Antwortdaten erwartet
        if (count > (17 + 81)) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        // Namen uebergehen
        count = 0;
        std::string lName(reinterpret_cast<char *>(&m_packet_data[count]));
        count += static_cast<int>( lName.size() ) + 1;

        // OK
        val = reinterpret_cast<char *>(&m_packet_data[count]);
    
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[17 + 81];
        memset(tmp, 0, sizeof(tmp));
        count = static_cast<int>( strlen( strncpy( (char*)tmp, name.c_str(), sizeof( tmp ) - 1 ) ) ) + 2;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_VARIABLE, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, name.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::write_memo_to_display( unsigned char device_address, unsigned char maschine, unsigned char time, unsigned char font, unsigned char audio, short flags, const std::string &text )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i, %i, %i, 0x%04X, %s ]\n", __FILE__, __LINE__, __func__, device_address, maschine, time, font, audio, flags, text.c_str());

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
// Maschinennummer (1 Byte)
// Time ist die Anzeigedauer. (1 Byte) (0 im Hindergrund uebernehmen, -1 staendig anzeigen)
// Schriftgroessennummer (1 Byte)
// Akkustisches Signal (1 Byte)
// 16bit reserve (2 Byte)
// Flags z.b. 0x0001 (2 Byte) fuer Antiproportionale Schrift
// Infotext
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        m_packet_data[count] = maschine;
        count += sizeof(unsigned char);
        m_packet_data[count] = time;
        count += sizeof(unsigned char);
        m_packet_data[count] = font;
        count += sizeof(unsigned char);
        m_packet_data[count] = audio;
        count += sizeof(unsigned char);
        *reinterpret_cast<short *>(&m_packet_data[count]) = 0;
        count += sizeof(short);
        *reinterpret_cast<short *>(&m_packet_data[count]) = flags;
        count += sizeof(short);

        // Namen uebernehmen
        std::copy(text.cbegin(), text.cend(), &m_packet_data[count]);
        count += static_cast<int>( text.size() );
        m_packet_data[count] = 0; // Nullterminierung
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_INFO, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_INFO, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_INFO, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, text.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::access_control_identification( unsigned char device_address, const char *tm, const char *id, const char *pin, char *status, int &status_length )
{
    trace("%s(%i):INFO(3): %s [ %i, %s, %s, %s ]\n", __FILE__, __LINE__, __func__, device_address, tm, id, pin);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        std::string TM_( tm );
        std::string Id_( id );
        std::string Pin_( pin );
        std::string regTM("[0-3]\\d[01]");
        std::string regId("\\w{1,20}");
        std::string regPin("\\d{0,8}");

        // Format pruefen
        if ( !std::regex_match( TM_, std::regex( regTM ) ) 
            || !std::regex_match( Id_, std::regex( regId ) ) 
            || !std::regex_match( Pin_, std::regex( regPin ) ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Namen uebernehmen
        count = static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], tm ) ) ) + 1;
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], id ) ) ) + 1;
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], pin ) ) ) + 1;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, m_packet_data, count, IV_PENDING_TRIALS(get_connection().get_read_timeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, result );
            return result;
        }

        if (count == 0) 
        {
            trace("%s(%i):WARNING(1): %s [ Aktiver Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        // Wert von Status uebernehmen
        int i;
        for (i=0; (i < count) && (i < status_length); i++) 
        {
            status[i] = m_packet_data[i];
        }
        // Uebergebene Laenge mitteilen
        status_length = i - 1;
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::reply_access_control_identification( unsigned char device_address )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {	
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):ERROR(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, result);
    }

    return result;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param device_address Anzusprechende Geraetenummer.
*/
int Datafox::Communication::Module_application::access_control_online_action( unsigned char device_address, const char *tm, unsigned int mask, unsigned char type, unsigned char duration )
{
    trace("%s(%i):INFO(3): %s [ %i, 0x08X, 0x%08X, %u, %u ]\n", __FILE__, __LINE__, __func__, device_address, tm, mask, type, duration);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        std::string TM_( tm );
        std::string regTM("[0-3]\\d[01]");

        if ( !std::regex_match( TM_, std::regex( regTM ) ) ) 
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Nummer des Tuermodul uebernehmen
        count = static_cast<int>( strlen( strcpy( (char*)m_packet_data, tm ) ) ) + 1;
        // Weiteren Werte uebernehmen
        *(unsigned int*)&m_packet_data[count] = mask;
        count += sizeof(unsigned int);
        m_packet_data[count] = type;
        count += sizeof(unsigned char);
        m_packet_data[count] = duration;
        count += sizeof(unsigned char);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_ONLINE_ACTION, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_ONLINE_ACTION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_ONLINE_ACTION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, tm);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::notify( unsigned char device_address, unsigned char type, int priority, char *timestamp, char *params )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i, %s ]\n", __FILE__, __LINE__, __func__, device_address, type, priority, params);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Benachrichtigungstyp uebernehmen
        m_packet_data[count] = type;
        count += sizeof(unsigned char);
        // Prioritaet eintragen
        *(short*)&m_packet_data[count] = priority;
        count += sizeof(short);
        // Zeitstempel
        if (timestamp == NULL) 
        {
            time_t tm_time;
            struct tm *ptm;
            tm_time = time(NULL);
            ptm = localtime(&tm_time);
            count += sprintf((char*)&m_packet_data[count], "%02i%02i-%02i-%02i %02i:%02i:%02i:%03i", 
                (ptm->tm_year + 1900) / 100,
                (ptm->tm_year + 1900) % 100,
                ptm->tm_mon + 1,
                ptm->tm_mday,
                ptm->tm_hour,
                ptm->tm_min,
                ptm->tm_sec, 0) + 1;
        }
        else 
        {
            count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], timestamp ) ) ) + 1;
        }
        // Leerstring
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], params ) ) ) + 1;

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_NOTIFY, m_packet_data, count)) != IV_ERROR_SUCCESS)
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_NOTIFY, result );
            return result;
        }

        // OK
        if (count > (int)sizeof(int)) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        unsigned char tmp[4];
        count = sizeof(tmp);
        tmp[0] = 1;

        switch (type)
        {
        case IV_VALUE_NOTIFY_RECORD:
            {
            CNotifyRecord nr;

            // Hat das Kanalobjekt den Aktiven Zustand?
            if (get_active_mode_state() < ams_active)
            {
                // Funktion derzeit nicht ausführbar.
                return 0x0001;
            }
            //TRACE("ActiveState: %i\n", GetActiveState());

            if (m_call_back_notify_record_available)
            {
                int pos = 0;

                // Aktuelle Kanalnummer eintragen
                nr.channelID = get_connection_id();

                // Aktuelle RemoteBusnummer eintragen
                nr.deviceID = get_device_address_receiver();

                // Meldungstypen uebernehmen
                nr.type = type;
                pos += sizeof(unsigned char);

                // Meldungsprioritaet uebernehmen
                nr.priority = priority;
                pos += sizeof(unsigned short);

                // Zeitstempel der Meldungserzeugung uebernehmen
                pos += static_cast<int>( strlen( strncpy( nr.timestamp, timestamp, sizeof( nr.timestamp ) - 1 ) ) );

                // Aktuellen Zeitstempel fuer Nachrichtenannahme halten
                nr.dt = time(NULL);

                // Meldung in Warteschlange stellen.
                if (m_call_back_notify_record_available(nr) != IV_ERROR_SUCCESS)
                {
                    // Die Annahme der Aktiven Datensatzmeldung schlug fehl (ggf. nicht aktiv).
                    return 0x0002;
                }

                trace("%s(%i):INFO(1): %s [ %i ]\n", __FILE__, __LINE__, __func__, get_connection_id());
            }
            else
            {
                // Die Annahme der Aktiven Datensatzmeldung ist nicht aktiv.
                return 0x0002;
            }
            break;
            }

        default:
            trace("%s(%i):INFO(1): %s [ Unknown type %i ]\n", __FILE__, __LINE__, __func__, type);
            break;
        }


        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_NOTIFY, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::notify_record_available( unsigned char device_address, int priority )
{
    trace("%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, priority);

    return notify(device_address, IV_VALUE_NOTIFY_RECORD, priority, NULL, (char*)"");
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param device_address Anzusprechende Geraetenummer.
* @param type 
0		Standardschrift(Sicherstellung der Abwärtskompatibilität)
1		16 Pixel (7Anzeigezeilen)
2		16 Pixel,feste breite(7Zeilen)
3		19 Pixel(6Anzeigezeilen)
4		19 Pixel,feste breite(6Zeilen)
5		21 Pixel(5Anzeigezeilen)
6		21 Pixel,feste breite(5Zeilen)
*/
int Datafox::Communication::Module_application::write_font_type( unsigned char device_address, unsigned char type )
{
    trace("%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, type);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Typewert uebernehmen
        count = sizeof(unsigned char);
        // Wert uebernehmen
        *m_packet_data = type;

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_TYPE_FONT, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_SET_TYPE_FONT, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_TYPE_FONT, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
* @brief Ermittelt die Anzahl der vorliegenden Datensaetze oder verwendetes Speichervolumen.
*
* @param [in]     device_address Anzusprechende Geraetenummer.
* @param [out]    number Parameter der die Anzahl Datensaetze erhaelt. Er wird auf -1 gesetzt, wenn der Wert nicht ermittelt werden konnte.
* @param [out]    memoryUsage Parameter der das Datenvolumen erhaelt. Er wird auf -1 gesetzt, wenn der Wert nicht ermittelt werden konnte.
* @param [out]    memorySize Parameter der das Gesamtvolumen fuer Datensaetze erhaelt. Er wird auf -1 gesetzt, wenn der Wert nicht ermittelt werden konnte.
*/
int Datafox::Communication::Module_application::read_records_volume( unsigned char device_address, int &number, int &memory_usage, int &memory_size )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, number, memory_usage, memory_size);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Zuerst wird davon ausgegangen, dass beide Werte nicht ermittelt werden koennen.
        number = -1;
        memory_usage = -1;
        memory_size = -1;

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_RECORD_VOLUME, m_packet_data, count)) == IV_ERROR_SUCCESS)
        {
            // Antwortdaten erwartet
            if (count != (int)sizeof(unsigned int)*3)
            {
                trace("%s(%i):WARNING(1): %s [ Ermittlung Anzahl Datensaetze und Speicher mit %i Byte, ungleich erwartete %i Byte ]\n", __FILE__, __LINE__, __func__, count - 3 * sizeof(int), 3 * sizeof(int));
            }

            // Anzahl der Datensaetze uebernehmen.
            number = *(int*)m_packet_data;
            if (count > (int)sizeof(int))
                memory_usage = *(int*)&m_packet_data[sizeof(int)];
            if (count > (int)sizeof(int)*2)
                memory_size = *(int*)&m_packet_data[sizeof(int)*2];

            return result;
        }

        // Nur der Modulfehler -Funktion unbekannt- wird akzeptiert.
        if (result < 0 || (result > 0 && ((result >> 8) & 0xFF) != 2)) {

            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_RECORD_VOLUME, result );
            return result;

        }

        // Wird die Funktion nicht unterstuetzt, dann wird versucht ueber die Systemvariable SYSTEM.STATUS die Information
        // zu ermitteln.
        std::string status;
        if ((result = read_system_variable(get_device_address_receiver(), "SYSTEM.STATUS", status)) != IV_ERROR_SUCCESS) {

            return result;
        }

        // Im Status wird nach "RECORDS:" gesucht.
        // 1. Es liegt nicht vor.
        // 2. Es liegt mit der Angabe des benoetigten Datenvolumens vor.
        // 3. Es liegt mit der Angabe der Datensatzanzahl und des Datenvolumens vor.
        // 4. Bei aktueller Berechnung (Zwecks MMC) kann auch "RECORDS: ?" kommen, dann wird noch ermittelt.
        char *pos, *end, *res;
        int val1 = -1; // Haelt den ersten ermittelten Wert.
        int val2 = -1; // Haelt den zweiten ermittelten Wert
        pos = strstr(&status[0], "RECORDS:");
        if (pos == NULL) {

            // Die Informationen konnten nicht ermittelt werden.
            return result;
        }
        pos += 9;
        end = strstr(pos, "\r");
        if (end == NULL) end = &status[strlen(&status[0])];
        if ((end - pos) > 20 || (pos - &status[0]) < 20) {

            // Die Informationen konnten nicht ausgewertet werden.
            return result;
        }
        if (*pos == '?') {

            // Die Information wird intern noch ermittelt und steht zur Zeit noch nicht zur Verfuegung.
            return result;
        }
        // Die vorderen 20 Zeichen des status-Puffers fuer halten der Werte verwenden.
        res = &status[0];
        memset(res, 0, 20);
        while (pos != end) {

            if (isdigit(*pos)) {
                *res = *pos;
                res++;
            } else if (*pos == ' ') {
                // Wert 1 oder 2 fuellen.
                if (val1 == -1) {
                    val1 = atoi(&status[0]);
                    res = &status[0];
                    memset(res, 0, 20);
                } else if (val2 == -1) {
                    val2 = atoi(&status[0]);
                    break;
                }
            } else if (*pos == 'b') {
                // Es liegt nur der Erste Wert als Volumeangabe vor.
                break;
            }

            // Naechstes Zeichen.
            pos++;
        }

        if (val2 == -1) {
            // Es liegt die Angabe in Datenvolumen vor.
            memory_usage = val1;
        } else {
            // Es liegen beide Angaben vor. Datensatzanzahl und volumen.
            number = val1;
            memory_usage = val2;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[64];
        count = sizeof(int)*3;
        *(int*)tmp = '?';
        *(int*)&tmp[sizeof(int)] = 0;
        *(int*)&tmp[sizeof(int)*2] = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_RECORD_VOLUME, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::access_control_identification_ext( unsigned char device_address, int master_id, const char *tm, const char *id, const char *pin, char *status, int &status_length )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %s, %s, %s ]\n", __FILE__, __LINE__, __func__, device_address, master_id, tm, id, pin);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        int res;
        std::string buffer;
        std::string TM_( tm );
        std::string Id_( id );
        std::string Pin_( pin );
        std::string regTM("[0-3]\\d[01]");
        std::string regId("\\w{1,20}");
        std::string regPin("\\d{0,8}");

        // Format pruefen
        if ( !std::regex_match( TM_, std::regex( regTM ) ) 
            || !std::regex_match( Id_, std::regex( regId ) ) 
            || !std::regex_match( Pin_, std::regex( regPin ) ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Angeggebene MasterId gegen die eingestellte prüfen.

        // Systemvariable ACCESS.MASTERID ermitteln.
        if ( (res = read_system_variable( device_address, "ACCESS.MASTERID", buffer )) != IV_ERROR_SUCCESS )
            return res;

        std::string currentMasterId( buffer );
        // Die anzusprechende MasterId muss größer oder gleich der aktuell eingestellten sein.
        if ( master_id < std::stoi( currentMasterId ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Namen uebernehmen
        *reinterpret_cast<unsigned short*>( m_packet_data ) = master_id;
        count = sizeof(unsigned short);
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], tm ) ) ) + 1;
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], id ) ) ) + 1;
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], pin ) ) ) + 1;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, m_packet_data, count, IV_PENDING_TRIALS(get_connection().get_read_timeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, result );
            return result;
        }

        if (count == 0) 
        {
            trace("%s(%i):WARNING(1): %s [ Aktiver Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        // Wert von Status uebernehmen
        int i;
        for (i=0; (i < count) && (i < status_length); i++) 
        {
            status[i] = m_packet_data[i];
        }
        // Uebergebene Laenge mitteilen
        status_length = i - 1;
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::reply_access_control_identification_ext( unsigned char device_address )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    } 
    else 
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {	
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):ERROR(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, result);
    }

    return result;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param device_address Anzusprechende Geraetenummer.
*/
int Datafox::Communication::Module_application::access_control_online_action_ext( unsigned char device_address, int master_id, const char *tm, unsigned int mask, unsigned char type, unsigned char duration )
{
    trace("%s(%i):INFO(3): %s [ %i, 0x08X, 0x%08X, %u, %u ]\n", __FILE__, __LINE__, __func__, device_address, tm, mask, type, duration);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        int res;
        std::string buffer;
        std::string TM_( tm );
        std::string regTM("[0-3]\\d[01]");

        if ( !std::regex_match( TM_, std::regex( regTM ) ) ) 
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Angeggebene MasterId gegen die eingestellte prüfen.

        // Systemvariable ACCESS.MASTERID ermitteln.
        if ( (res = read_system_variable( device_address, "ACCESS.MASTERID", buffer )) != IV_ERROR_SUCCESS )
            return res;

        std::string currentMasterId( buffer );
        // Die anzusprechende MasterId muss größer oder gleich der aktuell eingestellten sein.
        if ( master_id < std::stoi( currentMasterId ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // anzusprechende MasterId
        *reinterpret_cast<unsigned short*>( m_packet_data ) = master_id;
        count = sizeof(unsigned short);
        // Nummer des Tuermodul uebernehmen
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], tm ) ) ) + 1;
        // Weiteren Werte uebernehmen
        *(unsigned int*)&m_packet_data[count] = mask;
        count += sizeof(unsigned int);
        m_packet_data[count] = type;
        count += sizeof(unsigned char);
        m_packet_data[count] = duration;
        count += sizeof(unsigned char);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, tm);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::access_control_knob_command( unsigned char device_address, unsigned char knob_command, unsigned short master_id, unsigned char knob_id, const char *params, unsigned char *data, int *size_data )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i, %i, %s ]\n", __FILE__, __LINE__, __func__, device_address, knob_command, master_id, knob_id, params);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);

        // Namen uebernehmen
        *reinterpret_cast<unsigned char*>( &m_packet_data[count] ) = knob_command;
        count = sizeof(unsigned char);
        *reinterpret_cast<unsigned short*>( &m_packet_data[count] ) = master_id;
        count += sizeof(unsigned short);
        *reinterpret_cast<unsigned char*>( &m_packet_data[count] ) = knob_id;
        count += sizeof(unsigned char);
        count += static_cast<int>( strlen( strcpy( (char*)&m_packet_data[count], params ) ) ) + 1;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, m_packet_data, count, IV_PENDING_TRIALS(get_connection().get_read_timeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, result );
            return result;
        }

        // Die Daten werden als eine Listen von Hexwerten mit \r getrennt geliefert.
        // Keine Antwortdaten erwartet
        if ( knob_command == IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST && count != 64 ) 
        {
            trace( "%s(%i):WARNING(1): %s [ Es wurden 64 Byte Antwort erwartet und %i Byte erhalten. ]\n", __FILE__, __LINE__, __func__, count );
        }
        
        if ( knob_command == IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST )
        {
            // Die Daten müssen entsprechend aufbereitet werden.
            auto swapByte = []( unsigned char *d, unsigned char *s )
            {
                unsigned char tmp = *d;
                *d = *s;
                *s = tmp;
            };

            auto toLSB = [=]( unsigned char *p ) 
            { 
                swapByte( &p[0], &p[3] );
                swapByte( &p[1], &p[2] );
                return *reinterpret_cast<int *>( p );
            };

            // Ggf. gelieferte Daten übernehmen.
            for ( int i = 0; i < (count / 4) && ((i + 1) * 4 + i) < *size_data; ++i )
                sprintf( reinterpret_cast<char *>( &data[i * 9] ), "%08X%c", toLSB( &m_packet_data[i * 4] ), (i + 1) == (count / 4) ? '\0' : '\r' );

            *size_data = count;
        }
        else
        {
            // Daten wie sie sind übernehmen.
            int i = 0;
            for ( ; i < count && i < *size_data; ++i )
                data[i] = m_packet_data[i];

            *size_data = i;
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::reply_access_control_knob_command( unsigned char device_address )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {	
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):ERROR(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, result);
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::custom_command( unsigned char device_address, unsigned int identification, unsigned int custom_command, const char *data_in, int size_in, char *data_out, int *size_out )
{
    trace("%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, device_address, custom_command);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        set_device_address_receiver(device_address);
        
		*reinterpret_cast<unsigned int*>( &m_packet_data[count] ) = identification;
		count = sizeof(unsigned int);
		*reinterpret_cast<unsigned int*>( &m_packet_data[count] ) = custom_command;
		count += sizeof(unsigned int);

		*reinterpret_cast<unsigned int*>( &m_packet_data[count] ) = size_in;
		count += sizeof(unsigned int);

		for ( int i = 0; i < size_in; i ++ )
		{
			*reinterpret_cast<char*>( &m_packet_data[count] ) =  data_in[i];
			count += sizeof( char );
		}

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND, m_packet_data, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK, m_packet_data, count, IV_PENDING_TRIALS(get_connection().get_read_timeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK,result );
            return result;
        }

        // Die Daten werden als eine Listen von Hexwerten mit \r getrennt geliefert.
        // Keine Antwortdaten erwartet

        // Daten wie sie sind übernehmen.
        int i = 0;
        for ( ; i < count && i < *size_out; ++i )
            data_out[i] = m_packet_data[i];

        *size_out = i;
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        unsigned char tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param device_address Anzusprechende Geraetenummer.
 */
int Datafox::Communication::Module_application::reply_custom_command( unsigned char device_address )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, device_address);

    int result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        unsigned char tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):ERROR(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, result);
    }

    return result;
}
